@@ -1,9 +1,7 @@
-# coding=utf-8
-#
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2019 David R. MacIver
+# Most of this work is copyright (C) 2013-2020 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -15,8 +13,6 @@
 #
 # END HEADER
 
-from __future__ import absolute_import, division, print_function
-
 from collections import defaultdict
 from enum import IntEnum
 
@@ -26,12 +22,8 @@ from hypothesis.errors import Frozen, InvalidArgument, StopTest
 from hypothesis.internal.compat import (
     benchmark_time,
     bit_length,
-    hbytes,
-    hrange,
     int_from_bytes,
     int_to_bytes,
-    text_type,
-    unicode_safe_repr,
 )
 from hypothesis.internal.conjecture.junkdrawer import IntList, uniform
 from hypothesis.internal.conjecture.utils import calc_label_from_name
@@ -41,7 +33,7 @@ TOP_LABEL = calc_label_from_name("top")
 DRAW_BYTES_LABEL = calc_label_from_name("draw_bytes() in ConjectureData")
 
 
-class ExtraInformation(object):
+class ExtraInformation:
     """A class for holding shared state on a ``ConjectureData`` that should
     be added to the final ``ConjectureResult``."""
 
@@ -65,7 +57,7 @@ class Status(IntEnum):
 
 
 @attr.s(frozen=True, slots=True)
-class StructuralCoverageTag(object):
+class StructuralCoverageTag:
     label = attr.ib()
 
 
@@ -79,7 +71,7 @@ def structural_coverage(label):
         return STRUCTURAL_COVERAGE_CACHE.setdefault(label, StructuralCoverageTag(label))
 
 
-class Example(object):
+class Example:
     """Examples track the hierarchical structure of draws from the byte stream,
     within a single test run.
 
@@ -192,7 +184,7 @@ class Example(object):
         return [self.owner[i] for i in self.owner.children[self.index]]
 
 
-class ExampleProperty(object):
+class ExampleProperty:
     """There are many properties of examples that we calculate by
     essentially rerunning the test case multiple times based on the
     calls which we record in ExampleRecord.
@@ -293,7 +285,7 @@ STOP_EXAMPLE_NO_DISCARD_RECORD = 2
 START_EXAMPLE_RECORD = 3
 
 
-class ExampleRecord(object):
+class ExampleRecord:
     """Records the series of ``start_example``, ``stop_example``, and
     ``draw_bits`` calls so that these may be stored in ``Examples`` and
     replayed when we need to know about the structure of individual
@@ -330,7 +322,7 @@ class ExampleRecord(object):
         self.trail.append(DRAW_BITS_RECORD)
 
 
-class Examples(object):
+class Examples:
     """A lazy collection of ``Example`` objects, derived from
     the record of recorded behaviour in ``ExampleRecord``.
 
@@ -428,10 +420,24 @@ class Examples(object):
         def start_example(self, i, label_index):
             self.result[i] = label_index
 
+    @calculated_example_property
+    class mutator_groups(ExampleProperty):
+        def begin(self):
+            self.groups = defaultdict(list)
+
+        def start_example(self, i, label_index):
+            depth = len(self.example_stack)
+            self.groups[label_index, depth].append(i)
+
+        def finish(self):
+            # Discard groups with only one example, since the mutator can't
+            # do anything useful with them.
+            return [g for g in self.groups.values() if len(g) >= 2]
+
     @property
     def children(self):
         if self.__children is None:
-            self.__children = [IntList() for _ in hrange(len(self))]
+            self.__children = [IntList() for _ in range(len(self))]
             for i, p in enumerate(self.parentage):
                 if i > 0:
                     self.__children[p].append(i)
@@ -456,7 +462,7 @@ class Examples(object):
 
 
 @attr.s(slots=True, frozen=True)
-class Block(object):
+class Block:
     """Blocks track the flat list of lowest-level draws from the byte stream,
     within a single test run.
 
@@ -493,7 +499,7 @@ class Block(object):
         return self.forced or self.all_zero
 
 
-class Blocks(object):
+class Blocks:
     """A lazily calculated list of blocks for a particular ``ConjectureResult``
     or ``ConjectureData`` object.
 
@@ -649,12 +655,12 @@ class Blocks(object):
             self.owner = None
 
     def __iter__(self):
-        for i in hrange(len(self)):
+        for i in range(len(self)):
             yield self[i]
 
     def __repr__(self):
         parts = []
-        for i in hrange(len(self)):
+        for i in range(len(self)):
             b = self.__known_block(i)
             if b is None:
                 parts.append("...")
@@ -663,7 +669,7 @@ class Blocks(object):
         return "Block([%s])" % (", ".join(parts),)
 
 
-class _Overrun(object):
+class _Overrun:
     status = Status.OVERRUN
 
     def __repr__(self):
@@ -681,7 +687,7 @@ global_test_counter = 0
 MAX_DEPTH = 100
 
 
-class DataObserver(object):
+class DataObserver:
     """Observer class for recording the behaviour of a
     ConjectureData object, primarily used for tracking
     the behaviour in the tree cache."""
@@ -707,7 +713,7 @@ class DataObserver(object):
 
 
 @attr.s(slots=True)
-class ConjectureResult(object):
+class ConjectureResult:
     """Result class storing the parts of ConjectureData that we
     will care about after the original ConjectureData has outlived its
     usefulness."""
@@ -736,11 +742,11 @@ class ConjectureResult(object):
 
 # Masks for masking off the first byte of an n-bit buffer.
 # The appropriate mask is stored at position n % 8.
-BYTE_MASKS = [(1 << n) - 1 for n in hrange(8)]
+BYTE_MASKS = [(1 << n) - 1 for n in range(8)]
 BYTE_MASKS[0] = 255
 
 
-class ConjectureData(object):
+class ConjectureData:
     @classmethod
     def for_buffer(self, buffer, observer=None):
         return ConjectureData(
@@ -766,7 +772,7 @@ class ConjectureData(object):
         self.blocks = Blocks(self)
         self.buffer = bytearray()
         self.index = 0
-        self.output = u""
+        self.output = ""
         self.status = Status.VALID
         self.frozen = False
         global global_test_counter
@@ -779,7 +785,6 @@ class ConjectureData(object):
         self.draw_times = []
         self.max_depth = 0
         self.has_discards = False
-        self.consecutive_discard_counts = []
 
         self.__result = None
 
@@ -845,8 +850,8 @@ class ConjectureData(object):
 
     def note(self, value):
         self.__assert_not_frozen("note")
-        if not isinstance(value, text_type):
-            value = unicode_safe_repr(value)
+        if not isinstance(value, str):
+            value = repr(value)
         self.output += value
 
     def draw(self, strategy, label=None):
@@ -906,13 +911,11 @@ class ConjectureData(object):
         if self.depth > self.max_depth:
             self.max_depth = self.depth
         self.__example_record.start_example(label)
-        self.consecutive_discard_counts.append(0)
         self.labels_for_structure_stack.append({label})
 
     def stop_example(self, discard=False):
         if self.frozen:
             return
-        self.consecutive_discard_counts.pop()
         if discard:
             self.has_discards = True
         self.depth -= 1
@@ -927,23 +930,6 @@ class ConjectureData(object):
             else:
                 self.tags.update([structural_coverage(l) for l in labels_for_structure])
 
-        if self.consecutive_discard_counts:
-            # We block long sequences of discards. This helps us avoid performance
-            # problems where there is rejection sampling. In particular tests which
-            # have a very small actual state space but use rejection sampling will
-            # play badly with generate_novel_prefix() in DataTree, and will end up
-            # generating very long tests with long runs of the rejection sample.
-            if discard:
-                self.consecutive_discard_counts[-1] += 1
-                # 20 is a fairly arbitrary limit chosen mostly so that all of the
-                # existing tests passed under it. Essentially no reasonable
-                # generation should hit this limit when running in purely random
-                # mode, but unreasonable generation is fairly widespread, and our
-                # manipulation of the bitstream can make it more likely.
-                if self.consecutive_discard_counts[-1] > 20:
-                    self.mark_invalid()
-            else:
-                self.consecutive_discard_counts[-1] = 0
         if discard:
             # Once we've discarded an example, every test case starting with
             # this prefix contains discards. We prune the tree at that point so
@@ -983,7 +969,7 @@ class ConjectureData(object):
 
     def freeze(self):
         if self.frozen:
-            assert isinstance(self.buffer, hbytes)
+            assert isinstance(self.buffer, bytes)
             return
         self.finish_time = benchmark_time()
         assert len(self.buffer) == self.index
@@ -997,7 +983,7 @@ class ConjectureData(object):
 
         self.frozen = True
 
-        self.buffer = hbytes(self.buffer)
+        self.buffer = bytes(self.buffer)
         self.events = frozenset(self.events)
         self.observer.conclude_test(self.status, self.interesting_origin)
 
@@ -1030,7 +1016,7 @@ class ConjectureData(object):
         # If we have a number of bits that is not a multiple of 8
         # we have to mask off the high bits.
         buf[0] &= BYTE_MASKS[n % 8]
-        buf = hbytes(buf)
+        buf = bytes(buf)
         result = int_from_bytes(buf)
 
         self.observer.draw_bits(n, forced is not None, result)
@@ -1042,7 +1028,7 @@ class ConjectureData(object):
         self.index = len(self.buffer)
 
         if forced is not None:
-            self.forced_indices.update(hrange(initial, self.index))
+            self.forced_indices.update(range(initial, self.index))
 
         self.blocks.add_endpoint(self.index)
 
@@ -1056,7 +1042,7 @@ class ConjectureData(object):
     def write(self, string):
         """Write ``string`` to the output buffer."""
         self.__assert_not_frozen("write")
-        string = hbytes(string)
+        string = bytes(string)
         if not string:
             return
         self.draw_bits(len(string) * 8, forced=int_from_bytes(string))
