@@ -1,9 +1,7 @@
-# coding=utf-8
-#
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2019 David R. MacIver
+# Most of this work is copyright (C) 2013-2020 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -15,8 +13,6 @@
 #
 # END HEADER
 
-from __future__ import absolute_import, division, print_function
-
 import math
 
 from hypothesis.internal.conjecture.data import Status
@@ -26,7 +22,7 @@ from hypothesis.utils.dynamicvariables import DynamicVariable
 collector = DynamicVariable(None)
 
 
-class Statistics(object):
+class Statistics:
     def __init__(self, engine):
         self.passing_examples = len(engine.status_runtimes.get(Status.VALID, ()))
         self.invalid_examples = len(
@@ -34,6 +30,7 @@ class Statistics(object):
             + engine.status_runtimes.get(Status.OVERRUN, [])
         )
         self.failing_examples = len(engine.status_runtimes.get(Status.INTERESTING, ()))
+        self.targets = dict(engine.best_observed_targets)
 
         runtimes = sorted(
             engine.status_runtimes.get(Status.VALID, [])
@@ -73,8 +70,10 @@ class Statistics(object):
             )
 
         self.events = [
-            "%.2f%%, %s" % (c / engine.call_count * 100, e)
-            for e, c in sorted(engine.event_call_counts.items(), key=lambda x: -x[1])
+            "%6.2f%%, %s" % (c / engine.call_count * 100, e)
+            for e, c in sorted(
+                engine.event_call_counts.items(), key=lambda x: (-x[1], x[0])
+            )
         ]
 
         total_runtime = math.fsum(engine.all_runtimes)
@@ -106,12 +105,34 @@ class Statistics(object):
             % (self.draw_time_percentage,),
             "  - Stopped because %s" % (self.exit_reason,),
         ]
+        target_lines = describe_targets(self.targets)
+        if target_lines:
+            lines.append("  - " + target_lines[0])
+            lines.extend("    " + l for l in target_lines[1:])
         if self.events:
             lines.append("  - Events:")
             lines += ["    * %s" % (event,) for event in self.events]
         return lines
 
 
+def describe_targets(best_targets):
+    """Return a list of lines describing the results of `target`, if any."""
+    # These lines are included in the general statistics description below,
+    # but also printed immediately below failing examples to alleviate the
+    # "threshold problem" where shrinking can make severe bug look trival.
+    # See https://github.com/HypothesisWorks/hypothesis/issues/2180
+    if not best_targets:
+        return []
+    elif len(best_targets) == 1:
+        label, score = next(iter(best_targets.items()))
+        return ["Highest target score: {:g}  (label={!r})".format(score, label)]
+    else:
+        lines = ["Highest target scores:"]
+        for label, score in sorted(best_targets.items(), key=lambda x: x[::-1]):
+            lines.append("{:>16g}  (label={!r})".format(score, label))
+        return lines
+
+
 def note_engine_for_statistics(engine):
     callback = collector.value
     if callback is not None:
